__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "stagemodel"
__summary__ = "stagemodel: Stagewise models."
__uri__ = "https://github.com/ihmeuw-msca/stagemodel"

__version__ = "0.0.3"

__author__ = "IHME Math Sciences"
__email__ = "ihme.math.sciences@gmail.com"

__license__ = "BSD-2-Clause License"
__copyright__ = f"Copyright 2020 {__author__}"
